package shapecalculus;

import java.awt.Color;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import repast.simphony.context.Context;
import repast.simphony.space.SpatialException;
import repast.simphony.space.continuous.ContinuousSpace;
import repast.simphony.space.continuous.NdPoint;
import repast.simphony.space.grid.Grid;
import repast.simphony.space.grid.GridPoint;
import shapecalculus.behaviour.ActionDispatcher;
import shapecalculus.behaviour.AtomicBehaviour;
import shapecalculus.behaviour.Behaviour;
import shapecalculus.behaviour.NonAtomicBehaviour;
import shapecalculus.behaviour.channel.Channel;
import shapecalculus.physics.Motion;

public class BasicShape implements Shape {

	private Color color2d;
	private Color color3d;
	private double mass;
	private Behaviour behaviour;
	private NdPoint position, velocity;
	private GridPoint discretePosition;
	private GridPoint discreteVelocity;
	private Motion currentMotionLaw;
	private boolean behaviourPerformed;
	private Channel[] surfaces;
	private Context context;
	private ComplexShape complexShape;

	// private ContinuousSpace<Object> space;
	private Grid<Object> grid;
//	private ContinuousSpace<Object> space;

	public BasicShape(double mass, Context context, Grid<Object> grid,
			ContinuousSpace<Object> space, Behaviour behaviour) {
		this.mass = mass;
//		this.space = space;
		this.grid = grid;
		this.behaviour = behaviour;
		if (this.behaviour != null)
			this.behaviour.initPath(new ArrayList<Integer>());
		this.surfaces = new Channel[6];
		this.context = context;
		this.color2d = Color.blue;
		this.color3d = Color.blue;
	}

	public BasicShape(GridPoint discretePosition, GridPoint discreteVelocity,
			double mass, Context context, Grid<Object> grid,
			ContinuousSpace<Object> space, Behaviour behaviour) {
		this(mass, context, grid, space, behaviour);
		this.discretePosition = discretePosition;
		this.discreteVelocity = discreteVelocity;
		this.position = Utils.continuize(this.discretePosition);
		this.velocity = Utils.continuize(this.discreteVelocity);

	}

	public BasicShape(NdPoint position, NdPoint velocity, double mass,
			Context context, Grid<Object> grid, ContinuousSpace<Object> space,
			Behaviour behaviour) {
		this(mass, context, grid, space, behaviour);
		this.position = position;
		this.velocity = velocity;
		this.discretePosition = Utils.discretize(this.position);
		this.discreteVelocity = Utils.discretize(this.velocity);
	}

	public Context getContext() {
		return context;
	}

	public Channel getSurfaceAt(int direction) {
		if (direction < 6)
			return this.surfaces[direction];
		return null;
	}

	public void setSurfaceAt(Channel chan, int direction) {
		if (direction < 6)
			this.surfaces[direction] = chan;
	}

	public Behaviour getBehaviour() {
		return behaviour;
	}

	public void setBehaviour(Behaviour behaviour) {
		this.behaviour = behaviour;
		if (this.behaviour != null)
			this.behaviour.initPath(new ArrayList<Integer>());
	}

	@Override
	public void updateVelocity() {
		if (this.currentMotionLaw != null) {
			this.velocity = this.currentMotionLaw.calculateVelocity(this);
			this.discreteVelocity = Utils.discretize(this.velocity);
		}
		if (this.complexShape == null)
			move();
	}

	public boolean moveTo(NdPoint newPosition, boolean ignoreComplexShape) {
		GridPoint newDiscretePosition = Utils.discretize(newPosition);
		boolean moved=false;
		if (newPosition != this.position || newDiscretePosition != this.discretePosition) {
			if (!(this.complexShape == null || ignoreComplexShape)) {
				moved = this.getComplexShape().moveTo(
						Utils.sumPoints(newPosition,
								Utils.negPoint(this.position)));
			} else if (this.complexShape == null || ignoreComplexShape) {
//				if (moved=space.moveTo(this, Utils.getCoords(newPosition)))
//					this.position = newPosition;
				if (moved=grid.moveTo(this, Utils.getCoords(newDiscretePosition))){
					this.discretePosition = newDiscretePosition;
					this.position = newPosition;
				}
			}
		}
		return moved;
	}

	protected boolean move() {
		NdPoint newPosition = Utils.sumPoints(this.position, this.velocity);
		GridPoint newDiscretePosition = Utils.discretize(newPosition);
		boolean movedS = false;
		boolean movedG = false;
		boolean removed = false;
//		if (newPosition != this.position) {
//			try {
//				movedS = space.moveTo(this, Utils.getCoords(newPosition));
//				if (movedS) {
//					this.position = newPosition;
//				}
//			} catch (SpatialException e) {
//				removed = true;
//				this.context.remove(this);
//			}
//		}
		if (newDiscretePosition != this.discretePosition && !removed) {
			try {
				movedG = grid
						.moveTo(this, Utils.getCoords(newDiscretePosition));
				if (movedG) {
					this.position = newPosition;
					this.discretePosition = newDiscretePosition;
				}
			} catch (SpatialException e) {
				this.context.remove(this);
			}
		}
		return movedS || movedG;
	}

	@Override
	public void addMotion(Motion law) {
		if (law != null) {
			if (this.currentMotionLaw == null)
				this.currentMotionLaw = law;
			else {
				final Motion current = this.currentMotionLaw;
				final Motion newVel = law;
				this.currentMotionLaw = new Motion() {
					@Override
					public NdPoint calculateVelocity(Shape s) {
						NdPoint vel1 = current.calculateVelocity(s);
						NdPoint vel2 = newVel.calculateVelocity(s);
						return Utils.sumPoints(vel1, vel2,
								Utils.negPoint(s.getVelocity()));
					}
				};
			}
		}
	}

	@Override
	public void setMotion(Motion law) {
		this.currentMotionLaw = law;
	}

	@Override
	public Motion getMotion() {
		return this.currentMotionLaw;
	}

	public void dispatchEnabledActions(ActionDispatcher dispatcher) {
		if (this.behaviour != null) {
			if (this.behaviour instanceof NonAtomicBehaviour) {
				NonAtomicBehaviour b = (NonAtomicBehaviour) this.behaviour;
				Set<AtomicBehaviour> actions = b.getEnabledActions();
				for (AtomicBehaviour action : actions) {
					dispatcher.dispatchAction(action);
				}
			} else if (this.behaviour instanceof AtomicBehaviour) {
				AtomicBehaviour b = (AtomicBehaviour) this.behaviour;
				if (b.canPerform())
					dispatcher.dispatchAction(b);
			}
		}

	}

	@Override
	public double getMass() {
		return mass;
	}

	public void setMass(double mass) {
		this.mass = mass;
	}

	@Override
	public GridPoint getDiscretePosition() {
		return discretePosition;
	}

	@Override
	public NdPoint getPosition() {
		return this.position;
	}

	@Override
	public void setDiscretePosition(GridPoint discretePosition) {
		this.discretePosition = discretePosition;
		// this.position = Utils.continuize(discretePosition);
	}

	@Override
	public void setPosition(NdPoint position) {
		this.position = position;
		// this.discretePosition = Utils.discretize(position);
	}

	@Override
	public NdPoint getVelocity() {
		return this.velocity;
	}

	@Override
	public GridPoint getDiscreteVelocity() {
		return discreteVelocity;
	}

	@Override
	public void setVelocity(NdPoint velocity) {
		this.velocity = velocity;
		// this.discreteVelocity = Utils.discretize(this.velocity);
	}

	@Override
	public void setDiscreteVelocity(GridPoint discreteVelocity) {
		this.discreteVelocity = discreteVelocity;
		// this.velocity = Utils.continuize(discreteVelocity);
	}

	public int executeBehaviour(AtomicBehaviour action) {
		if (this.behaviour != null)
			return this.behaviour.perform(action.getPath().iterator());
		return -1;
	}

	@Override
	public void applyForce(NdPoint forceVector) {
		final NdPoint acceleration = new NdPoint(
				forceVector.getX() / this.mass, forceVector.getY() / this.mass,
				forceVector.getZ() / this.mass);
		Motion forceMotion = new Motion() {
			@Override
			public NdPoint calculateVelocity(Shape s) {
				return Utils.sumPoints(s.getVelocity(), acceleration);
			}
		};
		this.addMotion(forceMotion);
	}

	public boolean isBehaviourPerformed() {
		return behaviourPerformed;
	}

	public void setBehaviourPerformed(boolean behaviourPerformed) {
		this.behaviourPerformed = behaviourPerformed;
	}

	public List<BasicShape> buildComplexShape(List<BasicShape> tmp) {
		tmp.add(this);
		Channel c;
		BasicShape s;
		for (int i = 0; i < 6; i++) {
			c = this.surfaces[i];
			if (c != null) {
				s = c.getBindedTo().getShape();
				if (!tmp.contains(s))
					s.buildComplexShape(tmp);
			}
		}
		return tmp;
	}

	public void setComplexShape(ComplexShape complexShape) {
		this.complexShape = complexShape;
	}

	public ComplexShape getComplexShape() {
		return complexShape;
	}

	public Object clone() {
		BasicShape s = new BasicShape(position, velocity, mass, context, grid,
				null, (Behaviour) behaviour.clone());
		s.setMotion(this.currentMotionLaw);
		return s;
	}

	public Color getColor2d() {
		return color2d;
	}

	public void setColor2d(Color color2d) {
		this.color2d = color2d;
	}

	public Color getColor3d() {
		return color3d;
	}

	public void setColor3d(Color color3d) {
		this.color3d = color3d;
	}

}
